@@ -9,7 +9,8 @@ from django.http import HttpRequest, HttpResponse, HttpResponseNotAllowed
 from django.middleware.clickjacking import XFrameOptionsMiddleware
 from django.test import SimpleTestCase
 from django.utils.decorators import method_decorator
-from django.utils.functional import allow_lazy, lazy
+from django.utils.functional import keep_lazy, keep_lazy_text, lazy
+from django.utils.safestring import mark_safe
 from django.views.decorators.cache import (
     cache_control, cache_page, never_cache,
 )
@@ -25,6 +26,8 @@ from django.views.decorators.vary import vary_on_cookie, vary_on_headers
 def fully_decorated(request):
     """Expected __doc__"""
     return HttpResponse('<html><body>dummy</body></html>')
+
+
 fully_decorated.anything = "Expected __dict__"
 
 
@@ -67,8 +70,12 @@ full_decorator = compose(
     staff_member_required,
 
     # django.utils.functional
-    allow_lazy,
+    keep_lazy(HttpResponse),
+    keep_lazy_text,
     lazy,
+
+    # django.utils.safestring
+    mark_safe,
 )
 
 fully_decorated = full_decorator(fully_decorated)
@@ -78,8 +85,7 @@ class DecoratorsTest(TestCase):
 
     def test_attributes(self):
         """
-        Tests that django decorators set certain attributes of the wrapped
-        function.
+        Built-in decorators set certain attributes of the wrapped function.
         """
         self.assertEqual(fully_decorated.__name__, 'fully_decorated')
         self.assertEqual(fully_decorated.__doc__, 'Expected __doc__')
@@ -87,8 +93,7 @@ class DecoratorsTest(TestCase):
 
     def test_user_passes_test_composition(self):
         """
-        Test that the user_passes_test decorator can be applied multiple times
-        (#9474).
+        The user_passes_test decorator can be applied multiple times (#9474).
         """
         def test1(user):
             user.decorators_applied.append('test1')
@@ -104,10 +109,10 @@ class DecoratorsTest(TestCase):
         callback = user_passes_test(test1)(callback)
         callback = user_passes_test(test2)(callback)
 
-        class DummyUser(object):
+        class DummyUser:
             pass
 
-        class DummyRequest(object):
+        class DummyRequest:
             pass
 
         request = DummyRequest()
@@ -117,10 +122,7 @@ class DecoratorsTest(TestCase):
 
         self.assertEqual(response, ['test2', 'test1'])
 
-    def test_cache_page_new_style(self):
-        """
-        Test that we can call cache_page the new way
-        """
+    def test_cache_page(self):
         def my_view(request):
             return "response"
         my_view_cached = cache_page(123)(my_view)
@@ -157,6 +159,7 @@ def simple_dec(func):
         return func("test:" + arg)
     return wraps(func)(wrapper)
 
+
 simple_dec_m = method_decorator(simple_dec)
 
 
@@ -165,7 +168,8 @@ def myattr_dec(func):
     def wrapper(*args, **kwargs):
         return func(*args, **kwargs)
     wrapper.myattr = True
-    return wraps(func)(wrapper)
+    return wrapper
+
 
 myattr_dec_m = method_decorator(myattr_dec)
 
@@ -174,12 +178,13 @@ def myattr2_dec(func):
     def wrapper(*args, **kwargs):
         return func(*args, **kwargs)
     wrapper.myattr2 = True
-    return wraps(func)(wrapper)
+    return wrapper
+
 
 myattr2_dec_m = method_decorator(myattr2_dec)
 
 
-class ClsDec(object):
+class ClsDec:
     def __init__(self, myattr):
         self.myattr = myattr
 
@@ -195,7 +200,7 @@ class MethodDecoratorTests(SimpleTestCase):
     Tests for method_decorator
     """
     def test_preserve_signature(self):
-        class Test(object):
+        class Test:
             @simple_dec_m
             def say(self, arg):
                 return arg
@@ -204,16 +209,26 @@ class MethodDecoratorTests(SimpleTestCase):
 
     def test_preserve_attributes(self):
         # Sanity check myattr_dec and myattr2_dec
+        @myattr_dec
+        def func():
+            pass
+        self.assertIs(getattr(func, 'myattr', False), True)
+
+        @myattr2_dec
+        def func():
+            pass
+        self.assertIs(getattr(func, 'myattr2', False), True)
+
         @myattr_dec
         @myattr2_dec
         def func():
             pass
 
-        self.assertEqual(getattr(func, 'myattr', False), True)
-        self.assertEqual(getattr(func, 'myattr2', False), True)
+        self.assertIs(getattr(func, 'myattr', False), True)
+        self.assertIs(getattr(func, 'myattr2', False), False)
 
         # Decorate using method_decorator() on the method.
-        class TestPlain(object):
+        class TestPlain:
             @myattr_dec_m
             @myattr2_dec_m
             def method(self):
@@ -224,49 +239,71 @@ class MethodDecoratorTests(SimpleTestCase):
         # The decorators applied to the methods are applied before the ones
         # applied to the class.
         @method_decorator(myattr_dec_m, "method")
-        class TestMethodAndClass(object):
+        class TestMethodAndClass:
             @method_decorator(myattr2_dec_m)
             def method(self):
                 "A method"
                 pass
 
-        # Decorate using an iterable of decorators.
-        decorators = (myattr_dec_m, myattr2_dec_m)
-
-        @method_decorator(decorators, "method")
-        class TestIterable(object):
+        # Decorate using an iterable of function decorators.
+        @method_decorator((myattr_dec, myattr2_dec), 'method')
+        class TestFunctionIterable:
             def method(self):
                 "A method"
                 pass
 
-        for Test in (TestPlain, TestMethodAndClass, TestIterable):
-            self.assertEqual(getattr(Test().method, 'myattr', False), True)
-            self.assertEqual(getattr(Test().method, 'myattr2', False), True)
+        # Decorate using an iterable of method decorators.
+        decorators = (myattr_dec_m, myattr2_dec_m)
 
-            self.assertEqual(getattr(Test.method, 'myattr', False), True)
-            self.assertEqual(getattr(Test.method, 'myattr2', False), True)
+        @method_decorator(decorators, "method")
+        class TestMethodIterable:
+            def method(self):
+                "A method"
+                pass
 
-            self.assertEqual(Test.method.__doc__, 'A method')
-            self.assertEqual(Test.method.__name__, 'method')
+        tests = (TestPlain, TestMethodAndClass, TestFunctionIterable, TestMethodIterable)
+        for Test in tests:
+            with self.subTest(Test=Test):
+                self.assertIs(getattr(Test().method, 'myattr', False), True)
+                self.assertIs(getattr(Test().method, 'myattr2', False), True)
+                self.assertIs(getattr(Test.method, 'myattr', False), True)
+                self.assertIs(getattr(Test.method, 'myattr2', False), True)
+                self.assertEqual(Test.method.__doc__, 'A method')
+                self.assertEqual(Test.method.__name__, 'method')
+
+    def test_new_attribute(self):
+        """A decorator that sets a new attribute on the method."""
+        def decorate(func):
+            func.x = 1
+            return func
+
+        class MyClass:
+            @method_decorator(decorate)
+            def method(self):
+                return True
+
+        obj = MyClass()
+        self.assertEqual(obj.method.x, 1)
+        self.assertIs(obj.method(), True)
 
     def test_bad_iterable(self):
         decorators = {myattr_dec_m, myattr2_dec_m}
-        # The rest of the exception message differs between Python 2 and 3.
-        with self.assertRaisesMessage(TypeError, "'set' object"):
+        msg = "'set' object is not subscriptable"
+        with self.assertRaisesMessage(TypeError, msg):
             @method_decorator(decorators, "method")
-            class TestIterable(object):
+            class TestIterable:
                 def method(self):
                     "A method"
                     pass
 
     # Test for argumented decorator
     def test_argumented(self):
-        class Test(object):
+        class Test:
             @method_decorator(ClsDec(False))
             def method(self):
                 return True
 
-        self.assertEqual(Test().method(), False)
+        self.assertIs(Test().method(), False)
 
     def test_descriptors(self):
 
@@ -278,7 +315,7 @@ class MethodDecoratorTests(SimpleTestCase):
 
         method_dec = method_decorator(original_dec)
 
-        class bound_wrapper(object):
+        class bound_wrapper:
             def __init__(self, wrapped):
                 self.wrapped = wrapped
                 self.__name__ = wrapped.__name__
@@ -286,18 +323,18 @@ class MethodDecoratorTests(SimpleTestCase):
             def __call__(self, arg):
                 return self.wrapped(arg)
 
-            def __get__(self, instance, owner):
+            def __get__(self, instance, cls=None):
                 return self
 
-        class descriptor_wrapper(object):
+        class descriptor_wrapper:
             def __init__(self, wrapped):
                 self.wrapped = wrapped
                 self.__name__ = wrapped.__name__
 
-            def __get__(self, instance, owner):
-                return bound_wrapper(self.wrapped.__get__(instance, owner))
+            def __get__(self, instance, cls=None):
+                return bound_wrapper(self.wrapped.__get__(instance, cls))
 
-        class Test(object):
+        class Test:
             @method_dec
             @descriptor_wrapper
             def method(self, arg):
@@ -315,7 +352,7 @@ class MethodDecoratorTests(SimpleTestCase):
             return _wrapper
 
         @method_decorator(deco, name="method")
-        class Test(object):
+        class Test:
             def method(self):
                 return False
 
@@ -344,11 +381,11 @@ class MethodDecoratorTests(SimpleTestCase):
         decorators = (add_exclamation_mark, add_question_mark)
 
         @method_decorator(decorators, name="method")
-        class TestFirst(object):
+        class TestFirst:
             def method(self):
                 return "hello world"
 
-        class TestSecond(object):
+        class TestSecond:
             @method_decorator(decorators)
             def method(self):
                 return "hello world"
@@ -366,7 +403,7 @@ class MethodDecoratorTests(SimpleTestCase):
         )
         with self.assertRaisesMessage(TypeError, msg):
             @method_decorator(lambda: None, name="prop")
-            class Test(object):
+            class Test:
                 prop = 1
 
                 @classmethod
@@ -379,11 +416,11 @@ class MethodDecoratorTests(SimpleTestCase):
         """
         msg = (
             "The keyword argument `name` must be the name of a method of the "
-            "decorated class: <class 'Test'>. Got 'non_existing_method' instead"
+            "decorated class: <class 'Test'>. Got 'nonexistent_method' instead"
         )
         with self.assertRaisesMessage(ValueError, msg):
-            @method_decorator(lambda: None, name="non_existing_method")
-            class Test(object):
+            @method_decorator(lambda: None, name='nonexistent_method')
+            class Test:
                 @classmethod
                 def __module__(cls):
                     return "tests"
@@ -424,13 +461,13 @@ class XFrameOptionsDecoratorsTests(TestCase):
             return HttpResponse()
         req = HttpRequest()
         resp = a_view(req)
-        self.assertEqual(resp.get('X-Frame-Options', None), None)
+        self.assertIsNone(resp.get('X-Frame-Options', None))
         self.assertTrue(resp.xframe_options_exempt)
 
         # Since the real purpose of the exempt decorator is to suppress
         # the middleware's functionality, let's make sure it actually works...
         r = XFrameOptionsMiddleware().process_response(req, resp)
-        self.assertEqual(r.get('X-Frame-Options', None), None)
+        self.assertIsNone(r.get('X-Frame-Options', None))
 
 
 class NeverCacheDecoratorTest(TestCase):
@@ -441,5 +478,5 @@ class NeverCacheDecoratorTest(TestCase):
         r = a_view(HttpRequest())
         self.assertEqual(
             set(r['Cache-Control'].split(', ')),
-            {'max-age=0', 'no-cache', 'no-store', 'must-revalidate'},
+            {'max-age=0', 'no-cache', 'no-store', 'must-revalidate', 'private'},
         )
