@@ -1,15 +1,14 @@
-from __future__ import unicode_literals
-
 from math import ceil
 
 from django.db import IntegrityError, connection, models
+from django.db.models.deletion import Collector
 from django.db.models.sql.constants import GET_ITERATOR_CHUNK_SIZE
 from django.test import TestCase, skipIfDBFeature, skipUnlessDBFeature
-from django.utils.six.moves import range
 
 from .models import (
-    A, M, MR, R, S, T, Avatar, Base, Child, HiddenUser, HiddenUserProfile,
-    M2MFrom, M2MTo, MRNull, Parent, RChild, User, create_a, get_default_r,
+    MR, A, Avatar, Base, Child, HiddenUser, HiddenUserProfile, M, M2MFrom,
+    M2MTo, MRNull, Origin, Parent, R, RChild, RChildChild, Referrer, S, T,
+    User, create_a, get_default_r,
 )
 
 
@@ -22,6 +21,13 @@ class OnDeleteTests(TestCase):
         a.auto.delete()
         self.assertFalse(A.objects.filter(name='auto').exists())
 
+    def test_non_callable(self):
+        msg = 'on_delete must be callable.'
+        with self.assertRaisesMessage(TypeError, msg):
+            models.ForeignKey('self', on_delete=None)
+        with self.assertRaisesMessage(TypeError, msg):
+            models.OneToOneField('self', on_delete=None)
+
     def test_auto_nullable(self):
         a = create_a('auto_nullable')
         a.auto_nullable.delete()
@@ -63,7 +69,12 @@ class OnDeleteTests(TestCase):
 
     def test_protect(self):
         a = create_a('protect')
-        self.assertRaises(IntegrityError, a.protect.delete)
+        msg = (
+            "Cannot delete some instances of model 'R' because they are "
+            "referenced through a protected foreign key: 'A.protect'"
+        )
+        with self.assertRaisesMessage(IntegrityError, msg):
+            a.protect.delete()
 
     def test_do_nothing(self):
         # Testing DO_NOTHING is a bit harder: It would raise IntegrityError for a normal model,
@@ -82,7 +93,7 @@ class OnDeleteTests(TestCase):
 
     def test_do_nothing_qscount(self):
         """
-        Test that a models.DO_NOTHING relation doesn't trigger a query.
+        A models.DO_NOTHING relation doesn't trigger a query.
         """
         b = Base.objects.create()
         with self.assertNumQueries(1):
@@ -188,7 +199,7 @@ class DeletionTests(TestCase):
             obj = kwargs['instance']
             deleted.append(obj)
             if isinstance(obj, R):
-                related_setnull_sets.append(list(a.pk for a in obj.setnull_set.all()))
+                related_setnull_sets.append([a.pk for a in obj.setnull_set.all()])
 
         models.signals.pre_delete.connect(pre_delete)
         a = create_a('update_setnull')
@@ -224,12 +235,13 @@ class DeletionTests(TestCase):
         s2 = S.objects.create(pk=2, r=r)
         T.objects.create(pk=1, s=s1)
         T.objects.create(pk=2, s=s2)
+        RChild.objects.create(r_ptr=r)
         r.delete()
         self.assertEqual(
-            pre_delete_order, [(T, 2), (T, 1), (S, 2), (S, 1), (R, 1)]
+            pre_delete_order, [(T, 2), (T, 1), (RChild, 1), (S, 2), (S, 1), (R, 1)]
         )
         self.assertEqual(
-            post_delete_order, [(T, 1), (T, 2), (S, 1), (S, 2), (R, 1)]
+            post_delete_order, [(T, 1), (T, 2), (RChild, 1), (S, 1), (S, 2), (R, 1)]
         )
 
         models.signals.post_delete.disconnect(log_post_delete)
@@ -322,7 +334,7 @@ class DeletionTests(TestCase):
         # Calculate the number of queries needed.
         batch_size = connection.ops.bulk_batch_size(['pk'], objs)
         # The related fetches are done in batches.
-        batches = int(ceil(float(len(objs)) / batch_size))
+        batches = ceil(len(objs) / batch_size)
         # One query for Avatar.objects.all() and then one related fast delete for
         # each batch.
         fetches_to_mem = 1 + batches
@@ -339,12 +351,12 @@ class DeletionTests(TestCase):
 
         batch_size = max(connection.ops.bulk_batch_size(['pk'], range(TEST_SIZE)), 1)
 
-        # TEST_SIZE // batch_size (select related `T` instances)
+        # TEST_SIZE / batch_size (select related `T` instances)
         # + 1 (select related `U` instances)
-        # + TEST_SIZE // GET_ITERATOR_CHUNK_SIZE (delete `T` instances in batches)
+        # + TEST_SIZE / GET_ITERATOR_CHUNK_SIZE (delete `T` instances in batches)
         # + 1 (delete `s`)
-        expected_num_queries = (ceil(TEST_SIZE // batch_size) +
-                                ceil(TEST_SIZE // GET_ITERATOR_CHUNK_SIZE) + 2)
+        expected_num_queries = ceil(TEST_SIZE / batch_size)
+        expected_num_queries += ceil(TEST_SIZE / GET_ITERATOR_CHUNK_SIZE) + 2
 
         self.assertNumQueries(expected_num_queries, s.delete)
         self.assertFalse(S.objects.exists())
@@ -357,6 +369,25 @@ class DeletionTests(TestCase):
         self.assertFalse(RChild.objects.filter(id=child.id).exists())
         self.assertTrue(R.objects.filter(id=parent_id).exists())
 
+    def test_delete_with_keeping_parents_relationships(self):
+        child = RChild.objects.create()
+        parent_id = child.r_ptr_id
+        parent_referent_id = S.objects.create(r=child.r_ptr).pk
+        child.delete(keep_parents=True)
+        self.assertFalse(RChild.objects.filter(id=child.id).exists())
+        self.assertTrue(R.objects.filter(id=parent_id).exists())
+        self.assertTrue(S.objects.filter(pk=parent_referent_id).exists())
+
+        childchild = RChildChild.objects.create()
+        parent_id = childchild.rchild_ptr.r_ptr_id
+        child_id = childchild.rchild_ptr_id
+        parent_referent_id = S.objects.create(r=childchild.rchild_ptr.r_ptr).pk
+        childchild.delete(keep_parents=True)
+        self.assertFalse(RChildChild.objects.filter(id=childchild.id).exists())
+        self.assertTrue(RChild.objects.filter(id=child_id).exists())
+        self.assertTrue(R.objects.filter(id=parent_id).exists())
+        self.assertTrue(S.objects.filter(pk=parent_referent_id).exists())
+
     def test_queryset_delete_returns_num_rows(self):
         """
         QuerySet.delete() should return the number of deleted rows and a
@@ -424,6 +455,39 @@ class DeletionTests(TestCase):
         with self.assertNumQueries(2):
             avatar.delete()
 
+    def test_only_referenced_fields_selected(self):
+        """
+        Only referenced fields are selected during cascade deletion SELECT
+        unless deletion signals are connected.
+        """
+        origin = Origin.objects.create()
+        expected_sql = str(
+            Referrer.objects.only(
+                # Both fields are referenced by SecondReferrer.
+                'id', 'unique_field',
+            ).filter(origin__in=[origin]).query
+        )
+        with self.assertNumQueries(2) as ctx:
+            origin.delete()
+        self.assertEqual(ctx.captured_queries[0]['sql'], expected_sql)
+
+        def receiver(instance, **kwargs):
+            pass
+
+        # All fields are selected if deletion signals are connected.
+        for signal_name in ('pre_delete', 'post_delete'):
+            with self.subTest(signal=signal_name):
+                origin = Origin.objects.create()
+                signal = getattr(models.signals, signal_name)
+                signal.connect(receiver, sender=Referrer)
+                with self.assertNumQueries(2) as ctx:
+                    origin.delete()
+                self.assertIn(
+                    connection.ops.quote_name('large_field'),
+                    ctx.captured_queries[0]['sql'],
+                )
+                signal.disconnect(receiver, sender=Referrer)
+
 
 class FastDeleteTests(TestCase):
 
@@ -459,6 +523,14 @@ class FastDeleteTests(TestCase):
         self.assertEqual(User.objects.count(), 1)
         self.assertTrue(User.objects.filter(pk=u2.pk).exists())
 
+    def test_fast_delete_instance_set_pk_none(self):
+        u = User.objects.create()
+        # User can be fast-deleted.
+        collector = Collector(using='default')
+        self.assertTrue(collector.can_fast_delete(u))
+        u.delete()
+        self.assertIsNone(u.pk)
+
     def test_fast_delete_joined_qs(self):
         a = Avatar.objects.create(desc='a')
         User.objects.create(avatar=a)
@@ -473,11 +545,7 @@ class FastDeleteTests(TestCase):
         c = Child.objects.create()
         p = Parent.objects.create()
         # 1 for self, 1 for parent
-        # However, this doesn't work as child.parent access creates a query,
-        # and this means we will be generating extra queries (a lot for large
-        # querysets). This is not a fast-delete problem.
-        # self.assertNumQueries(2, c.delete)
-        c.delete()
+        self.assertNumQueries(2, c.delete)
         self.assertFalse(Child.objects.exists())
         self.assertEqual(Parent.objects.count(), 1)
         self.assertEqual(Parent.objects.filter(pk=p.pk).count(), 1)
