@@ -1,5 +1,3 @@
-from __future__ import unicode_literals
-
 import unittest
 from datetime import date, datetime, time, timedelta
 from decimal import Decimal
@@ -7,11 +5,10 @@ from operator import attrgetter, itemgetter
 from uuid import UUID
 
 from django.core.exceptions import FieldError
-from django.db import connection, models
-from django.db.models import F, Q, Max, Min, Sum, Value
+from django.db import models
+from django.db.models import F, Max, Min, Q, Sum, Value
 from django.db.models.expressions import Case, When
-from django.test import TestCase
-from django.utils import six
+from django.test import SimpleTestCase, TestCase
 
 from .models import CaseTestModel, Client, FKCaseTestModel, O2OCaseTestModel
 
@@ -275,6 +272,16 @@ class CaseExpressionTests(TestCase):
             [1, 4, 3, 3, 3, 2, 2]
         )
 
+    def test_annotate_with_empty_when(self):
+        objects = CaseTestModel.objects.annotate(
+            selected=Case(
+                When(pk__in=[], then=Value('selected')),
+                default=Value('not selected'), output_field=models.CharField()
+            )
+        )
+        self.assertEqual(len(objects), CaseTestModel.objects.count())
+        self.assertTrue(all(obj.selected == 'not selected' for obj in objects))
+
     def test_combined_expression(self):
         self.assertQuerysetEqual(
             CaseTestModel.objects.annotate(
@@ -289,12 +296,6 @@ class CaseExpressionTests(TestCase):
             transform=attrgetter('integer', 'test')
         )
 
-    if connection.vendor == 'sqlite' and connection.Database.sqlite_version_info < (3, 7, 0):
-        # There is a bug in sqlite < 3.7.0, where placeholder order is lost.
-        # Thus, the above query returns  <condition_value> + <result_value>
-        # for each matching case instead of <result_value> + 1 (#24148).
-        test_combined_expression = unittest.expectedFailure(test_combined_expression)
-
     def test_in_subquery(self):
         self.assertQuerysetEqual(
             CaseTestModel.objects.filter(
@@ -630,8 +631,6 @@ class CaseExpressionTests(TestCase):
     def test_update_binary(self):
         CaseTestModel.objects.update(
             binary=Case(
-                # fails on postgresql on Python 2.7 if output_field is not
-                # set explicitly
                 When(integer=1, then=Value(b'one', output_field=models.BinaryField())),
                 When(integer=2, then=Value(b'two', output_field=models.BinaryField())),
                 default=Value(b'', output_field=models.BinaryField()),
@@ -640,7 +639,7 @@ class CaseExpressionTests(TestCase):
         self.assertQuerysetEqual(
             CaseTestModel.objects.all().order_by('pk'),
             [(1, b'one'), (2, b'two'), (3, b''), (2, b'two'), (3, b''), (3, b''), (4, b'')],
-            transform=lambda o: (o.integer, six.binary_type(o.binary))
+            transform=lambda o: (o.integer, bytes(o.binary))
         )
 
     def test_update_boolean(self):
@@ -657,20 +656,6 @@ class CaseExpressionTests(TestCase):
             transform=attrgetter('integer', 'boolean')
         )
 
-    def test_update_comma_separated_integer(self):
-        CaseTestModel.objects.update(
-            comma_separated_integer=Case(
-                When(integer=1, then=Value('1')),
-                When(integer=2, then=Value('2,2')),
-                default=Value(''),
-            ),
-        )
-        self.assertQuerysetEqual(
-            CaseTestModel.objects.all().order_by('pk'),
-            [(1, '1'), (2, '2,2'), (3, ''), (2, '2,2'), (3, ''), (3, ''), (4, '')],
-            transform=attrgetter('integer', 'comma_separated_integer')
-        )
-
     def test_update_date(self):
         CaseTestModel.objects.update(
             date=Case(
@@ -763,7 +748,7 @@ class CaseExpressionTests(TestCase):
         self.assertQuerysetEqual(
             CaseTestModel.objects.all().order_by('pk'),
             [(1, '~/1'), (2, '~/2'), (3, ''), (2, '~/2'), (3, ''), (3, ''), (4, '')],
-            transform=lambda o: (o.integer, six.text_type(o.file))
+            transform=lambda o: (o.integer, str(o.file))
         )
 
     def test_update_file_path(self):
@@ -804,7 +789,7 @@ class CaseExpressionTests(TestCase):
         self.assertQuerysetEqual(
             CaseTestModel.objects.all().order_by('pk'),
             [(1, '~/1'), (2, '~/2'), (3, ''), (2, '~/2'), (3, ''), (3, ''), (4, '')],
-            transform=lambda o: (o.integer, six.text_type(o.image))
+            transform=lambda o: (o.integer, str(o.image))
         )
 
     def test_update_generic_ip_address(self):
@@ -835,6 +820,19 @@ class CaseExpressionTests(TestCase):
             transform=attrgetter('integer', 'null_boolean')
         )
 
+    def test_update_null_boolean_old(self):
+        CaseTestModel.objects.update(
+            null_boolean_old=Case(
+                When(integer=1, then=True),
+                When(integer=2, then=False),
+            ),
+        )
+        self.assertQuerysetEqual(
+            CaseTestModel.objects.all().order_by('pk'),
+            [(1, True), (2, False), (3, None), (2, False), (3, None), (3, None), (4, None)],
+            transform=attrgetter('integer', 'null_boolean_old')
+        )
+
     def test_update_positive_integer(self):
         CaseTestModel.objects.update(
             positive_integer=Case(
@@ -1262,6 +1260,15 @@ class CaseDocumentationExamples(TestCase):
             account_type=Client.PLATINUM,
             registered_on=date.today(),
         )
+        self.assertEqual(
+            Client.objects.aggregate(
+                regular=models.Count('pk', filter=Q(account_type=Client.REGULAR)),
+                gold=models.Count('pk', filter=Q(account_type=Client.GOLD)),
+                platinum=models.Count('pk', filter=Q(account_type=Client.PLATINUM)),
+            ),
+            {'regular': 2, 'gold': 1, 'platinum': 3}
+        )
+        # This was the example before the filter argument was added.
         self.assertEqual(
             Client.objects.aggregate(
                 regular=models.Sum(Case(
@@ -1279,3 +1286,59 @@ class CaseDocumentationExamples(TestCase):
             ),
             {'regular': 2, 'gold': 1, 'platinum': 3}
         )
+
+    def test_filter_example(self):
+        a_month_ago = date.today() - timedelta(days=30)
+        a_year_ago = date.today() - timedelta(days=365)
+        self.assertQuerysetEqual(
+            Client.objects.filter(
+                registered_on__lte=Case(
+                    When(account_type=Client.GOLD, then=a_month_ago),
+                    When(account_type=Client.PLATINUM, then=a_year_ago),
+                ),
+            ),
+            [('Jack Black', 'P')],
+            transform=attrgetter('name', 'account_type')
+        )
+
+    def test_hash(self):
+        expression_1 = Case(
+            When(account_type__in=[Client.REGULAR, Client.GOLD], then=1),
+            default=2,
+            output_field=models.IntegerField(),
+        )
+        expression_2 = Case(
+            When(account_type__in=(Client.REGULAR, Client.GOLD), then=1),
+            default=2,
+            output_field=models.IntegerField(),
+        )
+        expression_3 = Case(When(account_type__in=[Client.REGULAR, Client.GOLD], then=1), default=2)
+        expression_4 = Case(When(account_type__in=[Client.PLATINUM, Client.GOLD], then=2), default=1)
+        self.assertEqual(hash(expression_1), hash(expression_2))
+        self.assertNotEqual(hash(expression_2), hash(expression_3))
+        self.assertNotEqual(hash(expression_1), hash(expression_4))
+        self.assertNotEqual(hash(expression_3), hash(expression_4))
+
+
+class CaseWhenTests(SimpleTestCase):
+    def test_only_when_arguments(self):
+        msg = 'Positional arguments must all be When objects.'
+        with self.assertRaisesMessage(TypeError, msg):
+            Case(When(Q(pk__in=[])), object())
+
+    def test_invalid_when_constructor_args(self):
+        msg = (
+            'When() supports a Q object, a boolean expression, or lookups as '
+            'a condition.'
+        )
+        with self.assertRaisesMessage(TypeError, msg):
+            When(condition=object())
+        with self.assertRaisesMessage(TypeError, msg):
+            When(condition=Value(1, output_field=models.IntegerField()))
+        with self.assertRaisesMessage(TypeError, msg):
+            When()
+
+    def test_empty_q_object(self):
+        msg = "An empty Q() can't be used as a When() condition."
+        with self.assertRaisesMessage(ValueError, msg):
+            When(Q(), then=Value(True))
