@@ -27,8 +27,14 @@ class Article(models.Model):
         return self.headline
 
 
+class Country(models.Model):
+    id = models.SmallAutoField(primary_key=True)
+    name = models.CharField(max_length=50)
+
+
 class City(models.Model):
     id = models.BigAutoField(primary_key=True)
+    country = models.ForeignKey(Country, models.CASCADE, related_name='cities', null=True)
     name = models.CharField(max_length=50)
 
     def __str__(self):
@@ -70,6 +76,10 @@ class Child(models.Model):
     parent = models.ForeignKey(Parent, models.CASCADE)
 
 
+class ChildNullableParent(models.Model):
+    parent = models.ForeignKey(Parent, models.CASCADE, null=True)
+
+
 class ToFieldChild(models.Model):
     parent = models.ForeignKey(Parent, models.CASCADE, to_field='name', related_name='to_field_children')
 
