@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
 Testing using the Test Client
 
@@ -20,45 +19,28 @@ testing against the contexts and templates produced by a view,
 rather than the HTML rendered to the end-user.
 
 """
-from __future__ import unicode_literals
-
-import datetime
+import itertools
+import tempfile
+from unittest import mock
 
 from django.contrib.auth.models import User
 from django.core import mail
-from django.core.urlresolvers import reverse_lazy
 from django.http import HttpResponse
 from django.test import (
     Client, RequestFactory, SimpleTestCase, TestCase, override_settings,
 )
+from django.urls import reverse_lazy
 
-from .views import get_view, post_view, trace_view
+from .views import TwoArgException, get_view, post_view, trace_view
 
 
-@override_settings(PASSWORD_HASHERS=['django.contrib.auth.hashers.SHA1PasswordHasher'],
-                   ROOT_URLCONF='test_client.urls',)
+@override_settings(ROOT_URLCONF='test_client.urls')
 class ClientTest(TestCase):
 
     @classmethod
     def setUpTestData(cls):
-        cls.u1 = User.objects.create(
-            password='sha1$6efc0$f93efe9fd7542f25a7be94871ea45aa95de57161',
-            last_login=datetime.datetime(2006, 12, 17, 7, 3, 31), is_superuser=False, username='testclient',
-            first_name='Test', last_name='Client', email='testclient@example.com', is_staff=False, is_active=True,
-            date_joined=datetime.datetime(2006, 12, 17, 7, 3, 31)
-        )
-        cls.u2 = User.objects.create(
-            password='sha1$6efc0$f93efe9fd7542f25a7be94871ea45aa95de57161',
-            last_login=datetime.datetime(2006, 12, 17, 7, 3, 31), is_superuser=False, username='inactive',
-            first_name='Inactive', last_name='User', email='testclient@example.com', is_staff=False, is_active=False,
-            date_joined=datetime.datetime(2006, 12, 17, 7, 3, 31)
-        )
-        cls.u3 = User.objects.create(
-            password='sha1$6efc0$f93efe9fd7542f25a7be94871ea45aa95de57161',
-            last_login=datetime.datetime(2006, 12, 17, 7, 3, 31), is_superuser=False, username='staff',
-            first_name='Staff', last_name='Member', email='testclient@example.com', is_staff=True, is_active=True,
-            date_joined=datetime.datetime(2006, 12, 17, 7, 3, 31)
-        )
+        cls.u1 = User.objects.create_user(username='testclient', password='password')
+        cls.u2 = User.objects.create_user(username='inactive', password='password', is_active=False)
 
     def test_get_view(self):
         "GET a view"
@@ -72,6 +54,19 @@ class ClientTest(TestCase):
         self.assertEqual(response.context['var'], '\xf2')
         self.assertEqual(response.templates[0].name, 'GET Template')
 
+    def test_query_string_encoding(self):
+        # WSGI requires latin-1 encoded strings.
+        response = self.client.get('/get_view/?var=1\ufffd')
+        self.assertEqual(response.context['var'], '1\ufffd')
+
+    def test_get_data_none(self):
+        msg = (
+            "Cannot encode None for key 'value' in a query string. Did you "
+            "mean to pass an empty string or omit the value?"
+        )
+        with self.assertRaisesMessage(TypeError, msg):
+            self.client.get('/get_view/', {'value': None})
+
     def test_get_post_view(self):
         "GET a view that normally expects POSTs"
         response = self.client.get('/post_view/', {})
@@ -105,6 +100,53 @@ class ClientTest(TestCase):
         self.assertEqual(response.templates[0].name, 'POST Template')
         self.assertContains(response, 'Data received')
 
+    def test_post_data_none(self):
+        msg = (
+            "Cannot encode None for key 'value' as POST data. Did you mean "
+            "to pass an empty string or omit the value?"
+        )
+        with self.assertRaisesMessage(TypeError, msg):
+            self.client.post('/post_view/', {'value': None})
+
+    def test_json_serialization(self):
+        """The test client serializes JSON data."""
+        methods = ('post', 'put', 'patch', 'delete')
+        tests = (
+            ({'value': 37}, {'value': 37}),
+            ([37, True], [37, True]),
+            ((37, False), [37, False]),
+        )
+        for method in methods:
+            with self.subTest(method=method):
+                for data, expected in tests:
+                    with self.subTest(data):
+                        client_method = getattr(self.client, method)
+                        method_name = method.upper()
+                        response = client_method('/json_view/', data, content_type='application/json')
+                        self.assertEqual(response.status_code, 200)
+                        self.assertEqual(response.context['data'], expected)
+                        self.assertContains(response, 'Viewing %s page.' % method_name)
+
+    def test_json_encoder_argument(self):
+        """The test Client accepts a json_encoder."""
+        mock_encoder = mock.MagicMock()
+        mock_encoding = mock.MagicMock()
+        mock_encoder.return_value = mock_encoding
+        mock_encoding.encode.return_value = '{"value": 37}'
+
+        client = self.client_class(json_encoder=mock_encoder)
+        # Vendored tree JSON content types are accepted.
+        client.post('/json_view/', {'value': 37}, content_type='application/vnd.api+json')
+        self.assertTrue(mock_encoder.called)
+        self.assertTrue(mock_encoding.encode.called)
+
+    def test_put(self):
+        response = self.client.put('/put_view/', {'foo': 'bar'})
+        self.assertEqual(response.status_code, 200)
+        self.assertEqual(response.templates[0].name, 'PUT Template')
+        self.assertEqual(response.context['data'], "{'foo': 'bar'}")
+        self.assertEqual(response.context['Content-Length'], '14')
+
     def test_trace(self):
         """TRACE a view"""
         response = self.client.trace('/trace_view/')
@@ -120,9 +162,8 @@ class ClientTest(TestCase):
 
     def test_response_attached_request(self):
         """
-        Check that the returned response has a ``request`` attribute with the
-        originating environ dict and a ``wsgi_request`` with the originating
-        ``WSGIRequest`` instance.
+        The returned response has a ``request`` attribute with the originating
+        environ dict and a ``wsgi_request`` with the originating WSGIRequest.
         """
         response = self.client.get("/header_view/")
 
@@ -160,8 +201,7 @@ class ClientTest(TestCase):
         test_doc = """<?xml version="1.0" encoding="utf-8"?>
         <library><book><title>Blink</title><author>Malcolm Gladwell</author></book></library>
         """
-        response = self.client.post("/raw_post_view/", test_doc,
-                                    content_type="text/xml")
+        response = self.client.post('/raw_post_view/', test_doc, content_type='text/xml')
         self.assertEqual(response.status_code, 200)
         self.assertEqual(response.templates[0].name, "Book template")
         self.assertEqual(response.content, b"Blink - Malcolm Gladwell")
@@ -181,34 +221,34 @@ class ClientTest(TestCase):
     def test_redirect(self):
         "GET a URL that redirects elsewhere"
         response = self.client.get('/redirect_view/')
-        # Check that the response was a 302 (redirect)
         self.assertRedirects(response, '/get_view/')
 
     def test_redirect_with_query(self):
         "GET a URL that redirects with given GET parameters"
         response = self.client.get('/redirect_view/', {'var': 'value'})
-
-        # Check if parameters are intact
         self.assertRedirects(response, '/get_view/?var=value')
 
+    def test_redirect_with_query_ordering(self):
+        """assertRedirects() ignores the order of query string parameters."""
+        response = self.client.get('/redirect_view/', {'var': 'value', 'foo': 'bar'})
+        self.assertRedirects(response, '/get_view/?var=value&foo=bar')
+        self.assertRedirects(response, '/get_view/?foo=bar&var=value')
+
     def test_permanent_redirect(self):
         "GET a URL that redirects permanently elsewhere"
         response = self.client.get('/permanent_redirect_view/')
-        # Check that the response was a 301 (permanent redirect)
         self.assertRedirects(response, '/get_view/', status_code=301)
 
     def test_temporary_redirect(self):
         "GET a URL that does a non-permanent redirect"
         response = self.client.get('/temporary_redirect_view/')
-        # Check that the response was a 302 (non-permanent redirect)
         self.assertRedirects(response, '/get_view/', status_code=302)
 
     def test_redirect_to_strange_location(self):
         "GET a URL that redirects to a non-200 page"
         response = self.client.get('/double_redirect_view/')
-
-        # Check that the response was a 302, and that
-        # the attempt to get the redirection location returned 301 when retrieved
+        # The response was a 302, and that the attempt to get the redirection
+        # location returned 301 when retrieved
         self.assertRedirects(response, '/permanent_redirect_view/', target_status_code=301)
 
     def test_follow_redirect(self):
@@ -217,6 +257,51 @@ class ClientTest(TestCase):
         self.assertRedirects(response, '/get_view/', status_code=302, target_status_code=200)
         self.assertEqual(len(response.redirect_chain), 2)
 
+    def test_follow_relative_redirect(self):
+        "A URL with a relative redirect can be followed."
+        response = self.client.get('/accounts/', follow=True)
+        self.assertEqual(response.status_code, 200)
+        self.assertEqual(response.request['PATH_INFO'], '/accounts/login/')
+
+    def test_follow_relative_redirect_no_trailing_slash(self):
+        "A URL with a relative redirect with no trailing slash can be followed."
+        response = self.client.get('/accounts/no_trailing_slash', follow=True)
+        self.assertEqual(response.status_code, 200)
+        self.assertEqual(response.request['PATH_INFO'], '/accounts/login/')
+
+    def test_follow_307_and_308_redirect(self):
+        """
+        A 307 or 308 redirect preserves the request method after the redirect.
+        """
+        methods = ('get', 'post', 'head', 'options', 'put', 'patch', 'delete', 'trace')
+        codes = (307, 308)
+        for method, code in itertools.product(methods, codes):
+            with self.subTest(method=method, code=code):
+                req_method = getattr(self.client, method)
+                response = req_method('/redirect_view_%s/' % code, data={'value': 'test'}, follow=True)
+                self.assertEqual(response.status_code, 200)
+                self.assertEqual(response.request['PATH_INFO'], '/post_view/')
+                self.assertEqual(response.request['REQUEST_METHOD'], method.upper())
+
+    def test_follow_307_and_308_preserves_post_data(self):
+        for code in (307, 308):
+            with self.subTest(code=code):
+                response = self.client.post('/redirect_view_%s/' % code, data={'value': 'test'}, follow=True)
+                self.assertContains(response, 'test is the value')
+
+    def test_follow_307_and_308_preserves_put_body(self):
+        for code in (307, 308):
+            with self.subTest(code=code):
+                response = self.client.put('/redirect_view_%s/?to=/put_view/' % code, data='a=b', follow=True)
+                self.assertContains(response, 'a=b is the body')
+
+    def test_follow_307_and_308_preserves_get_params(self):
+        data = {'var': 30, 'to': '/get_view/'}
+        for code in (307, 308):
+            with self.subTest(code=code):
+                response = self.client.get('/redirect_view_%s/' % code, data=data, follow=True)
+                self.assertContains(response, '30 is the value')
+
     def test_redirect_http(self):
         "GET a URL that redirects to an http URI"
         response = self.client.get('/http_redirect_view/', follow=True)
@@ -230,8 +315,6 @@ class ClientTest(TestCase):
     def test_notfound_response(self):
         "GET a URL that responds as '404:Not Found'"
         response = self.client.get('/bad_view/')
-
-        # Check that the response was a 404, and that the content contains MAGIC
         self.assertContains(response, 'MAGIC', status_code=404)
 
     def test_valid_form(self):
@@ -256,7 +339,7 @@ class ClientTest(TestCase):
         response = self.client.get('/form_view/', data=hints)
         self.assertEqual(response.status_code, 200)
         self.assertTemplateUsed(response, "Form GET Template")
-        # Check that the multi-value data has been rolled out ok
+        # The multi-value data has been rolled out ok
         self.assertContains(response, 'Select a valid choice.', 0)
 
     def test_incomplete_data_form(self):
@@ -341,14 +424,14 @@ class ClientTest(TestCase):
         "GET an invalid URL"
         response = self.client.get('/unknown_view/')
 
-        # Check that the response was a 404
+        # The response was a 404
         self.assertEqual(response.status_code, 404)
 
     def test_url_parameters(self):
         "Make sure that URL ;-parameters are not stripped."
         response = self.client.get('/unknown_view/;some-parameter')
 
-        # Check that the path in the response includes it (ignore that it's a 404)
+        # The path in the response includes it (ignore that it's a 404)
         self.assertEqual(response.request['PATH_INFO'], '/unknown_view/;some-parameter')
 
     def test_view_with_login(self):
@@ -367,6 +450,13 @@ class ClientTest(TestCase):
         self.assertEqual(response.status_code, 200)
         self.assertEqual(response.context['user'].username, 'testclient')
 
+    @override_settings(
+        INSTALLED_APPS=['django.contrib.auth'],
+        SESSION_ENGINE='django.contrib.sessions.backends.file',
+    )
+    def test_view_with_login_when_sessions_app_is_not_installed(self):
+        self.test_view_with_login()
+
     def test_view_with_force_login(self):
         "Request a page that is protected with @login_required"
         # Get the page without logging in. Should result in 302.
@@ -451,11 +541,21 @@ class ClientTest(TestCase):
         self.assertFalse(login)
 
     def test_view_with_inactive_login(self):
-        "Request a page that is protected with @login, but use an inactive login"
+        """
+        An inactive user may login if the authenticate backend allows it.
+        """
+        credentials = {'username': 'inactive', 'password': 'password'}
+        self.assertFalse(self.client.login(**credentials))
 
-        login = self.client.login(username='inactive', password='password')
-        self.assertFalse(login)
+        with self.settings(AUTHENTICATION_BACKENDS=['django.contrib.auth.backends.AllowAllUsersModelBackend']):
+            self.assertTrue(self.client.login(**credentials))
 
+    @override_settings(
+        AUTHENTICATION_BACKENDS=[
+            'django.contrib.auth.backends.ModelBackend',
+            'django.contrib.auth.backends.AllowAllUsersModelBackend',
+        ]
+    )
     def test_view_with_inactive_force_login(self):
         "Request a page that is protected with @login, but use an inactive login"
 
@@ -464,7 +564,7 @@ class ClientTest(TestCase):
         self.assertRedirects(response, '/accounts/login/?next=/login_protected_view/')
 
         # Log in
-        self.client.force_login(self.u2)
+        self.client.force_login(self.u2, backend='django.contrib.auth.backends.AllowAllUsersModelBackend')
 
         # Request a page that requires a login
         response = self.client.get('/login_protected_view/')
@@ -529,6 +629,34 @@ class ClientTest(TestCase):
         self.assertEqual(response.status_code, 200)
         self.assertEqual(response.context['user'].username, 'testclient')
 
+    @override_settings(
+        AUTHENTICATION_BACKENDS=[
+            'django.contrib.auth.backends.ModelBackend',
+            'test_client.auth_backends.TestClientBackend',
+        ],
+    )
+    def test_force_login_without_backend(self):
+        """
+        force_login() without passing a backend and with multiple backends
+        configured should automatically use the first backend.
+        """
+        self.client.force_login(self.u1)
+        response = self.client.get('/login_protected_view/')
+        self.assertEqual(response.status_code, 200)
+        self.assertEqual(response.context['user'].username, 'testclient')
+        self.assertEqual(self.u1.backend, 'django.contrib.auth.backends.ModelBackend')
+
+    @override_settings(AUTHENTICATION_BACKENDS=[
+        'test_client.auth_backends.BackendWithoutGetUserMethod',
+        'django.contrib.auth.backends.ModelBackend',
+    ])
+    def test_force_login_with_backend_missing_get_user(self):
+        """
+        force_login() skips auth backends without a get_user() method.
+        """
+        self.client.force_login(self.u1)
+        self.assertEqual(self.u1.backend, 'django.contrib.auth.backends.ModelBackend')
+
     @override_settings(SESSION_ENGINE="django.contrib.sessions.backends.signed_cookies")
     def test_logout_cookie_sessions(self):
         self.test_logout()
@@ -586,34 +714,67 @@ class ClientTest(TestCase):
         response = self.client.get('/django_project_redirect/')
         self.assertRedirects(response, 'https://www.djangoproject.com/', fetch_redirect_response=False)
 
+    def test_external_redirect_with_fetch_error_msg(self):
+        """
+        assertRedirects without fetch_redirect_response=False raises
+        a relevant ValueError rather than a non-descript AssertionError.
+        """
+        response = self.client.get('/django_project_redirect/')
+        msg = (
+            "The test client is unable to fetch remote URLs (got "
+            "https://www.djangoproject.com/). If the host is served by Django, "
+            "add 'www.djangoproject.com' to ALLOWED_HOSTS. "
+            "Otherwise, use assertRedirects(..., fetch_redirect_response=False)."
+        )
+        with self.assertRaisesMessage(ValueError, msg):
+            self.assertRedirects(response, 'https://www.djangoproject.com/')
+
     def test_session_modifying_view(self):
         "Request a page that modifies the session"
         # Session value isn't set initially
-        try:
+        with self.assertRaises(KeyError):
             self.client.session['tobacconist']
-            self.fail("Shouldn't have a session value")
-        except KeyError:
-            pass
 
         self.client.post('/session_view/')
-
-        # Check that the session was modified
+        # The session was modified
         self.assertEqual(self.client.session['tobacconist'], 'hovercraft')
 
+    @override_settings(
+        INSTALLED_APPS=[],
+        SESSION_ENGINE='django.contrib.sessions.backends.file',
+    )
+    def test_sessions_app_is_not_installed(self):
+        self.test_session_modifying_view()
+
+    @override_settings(
+        INSTALLED_APPS=[],
+        SESSION_ENGINE='django.contrib.sessions.backends.nonexistent',
+    )
+    def test_session_engine_is_invalid(self):
+        with self.assertRaisesMessage(ImportError, 'nonexistent'):
+            self.test_session_modifying_view()
+
     def test_view_with_exception(self):
         "Request a page that is known to throw an error"
-        self.assertRaises(KeyError, self.client.get, "/broken_view/")
+        with self.assertRaises(KeyError):
+            self.client.get("/broken_view/")
 
-        # Try the same assertion, a different way
-        try:
-            self.client.get('/broken_view/')
-            self.fail('Should raise an error')
-        except KeyError:
-            pass
+    def test_exc_info(self):
+        client = Client(raise_request_exception=False)
+        response = client.get("/broken_view/")
+        self.assertEqual(response.status_code, 500)
+        exc_type, exc_value, exc_traceback = response.exc_info
+        self.assertIs(exc_type, KeyError)
+        self.assertIsInstance(exc_value, KeyError)
+        self.assertEqual(str(exc_value), "'Oops! Looks like you wrote some bad code.'")
+        self.assertIsNotNone(exc_traceback)
+
+    def test_exc_info_none(self):
+        response = self.client.get("/get_view/")
+        self.assertIsNone(response.exc_info)
 
     def test_mail_sending(self):
-        "Test that mail is redirected to a dummy outbox during test setup"
-
+        "Mail is redirected to a dummy outbox during test setup"
         response = self.client.get('/mail_sending_view/')
         self.assertEqual(response.status_code, 200)
 
@@ -625,16 +786,23 @@ class ClientTest(TestCase):
         self.assertEqual(mail.outbox[0].to[1], 'second@example.com')
 
     def test_reverse_lazy_decodes(self):
-        "Ensure reverse_lazy works in the test client"
+        "reverse_lazy() works in the test client"
         data = {'var': 'data'}
         response = self.client.get(reverse_lazy('get_view'), data)
 
         # Check some response details
         self.assertContains(response, 'This is a test')
 
-    def test_mass_mail_sending(self):
-        "Test that mass mail is redirected to a dummy outbox during test setup"
+    def test_relative_redirect(self):
+        response = self.client.get('/accounts/')
+        self.assertRedirects(response, '/accounts/login/')
 
+    def test_relative_redirect_no_trailing_slash(self):
+        response = self.client.get('/accounts/no_trailing_slash')
+        self.assertRedirects(response, '/accounts/login/')
+
+    def test_mass_mail_sending(self):
+        "Mass mail is redirected to a dummy outbox during test setup"
         response = self.client.get('/mass_mail_sending_view/')
         self.assertEqual(response.status_code, 200)
 
@@ -659,9 +827,25 @@ class ClientTest(TestCase):
         with self.assertRaisesMessage(Exception, 'exception message'):
             self.client.get('/nesting_exception_view/')
 
+    def test_response_raises_multi_arg_exception(self):
+        """A request may raise an exception with more than one required arg."""
+        with self.assertRaises(TwoArgException) as cm:
+            self.client.get('/two_arg_exception/')
+        self.assertEqual(cm.exception.args, ('one', 'two'))
+
+    def test_uploading_temp_file(self):
+        with tempfile.TemporaryFile() as test_file:
+            response = self.client.post('/upload_view/', data={'temp_file': test_file})
+        self.assertEqual(response.content, b'temp_file')
+
+    def test_uploading_named_temp_file(self):
+        test_file = tempfile.NamedTemporaryFile()
+        response = self.client.post('/upload_view/', data={'named_temp_file': test_file})
+        self.assertEqual(response.content, b'named_temp_file')
+
 
 @override_settings(
-    MIDDLEWARE_CLASSES=['django.middleware.csrf.CsrfViewMiddleware'],
+    MIDDLEWARE=['django.middleware.csrf.CsrfViewMiddleware'],
     ROOT_URLCONF='test_client.urls',
 )
 class CSRFEnabledClientTests(SimpleTestCase):
@@ -669,11 +853,9 @@ class CSRFEnabledClientTests(SimpleTestCase):
     def test_csrf_enabled_client(self):
         "A client can be instantiated with CSRF checks enabled"
         csrf_client = Client(enforce_csrf_checks=True)
-
         # The normal client allows the post
         response = self.client.post('/post_view/', {})
         self.assertEqual(response.status_code, 200)
-
         # The CSRF-enabled client rejects it
         response = csrf_client.post('/post_view/', {})
         self.assertEqual(response.status_code, 403)
@@ -688,10 +870,11 @@ class CustomTestClientTest(SimpleTestCase):
 
     def test_custom_test_client(self):
         """A test case can specify a custom class for self.client."""
-        self.assertEqual(hasattr(self.client, "i_am_customized"), True)
+        self.assertIs(hasattr(self.client, "i_am_customized"), True)
 
 
-_generic_view = lambda request: HttpResponse(status=200)
+def _generic_view(request):
+    return HttpResponse(status=200)
 
 
 @override_settings(ROOT_URLCONF='test_client.urls')
@@ -709,9 +892,7 @@ class RequestFactoryTest(SimpleTestCase):
         ('options', _generic_view),
         ('trace', trace_view),
     )
-
-    def setUp(self):
-        self.request_factory = RequestFactory()
+    request_factory = RequestFactory()
 
     def test_request_factory(self):
         """The request factory implements all the HTTP/1.1 methods."""
@@ -719,7 +900,6 @@ class RequestFactoryTest(SimpleTestCase):
             method = getattr(self.request_factory, method_name)
             request = method('/somewhere/')
             response = view(request)
-
             self.assertEqual(response.status_code, 200)
 
     def test_get_request_from_factory(self):
@@ -728,8 +908,6 @@ class RequestFactoryTest(SimpleTestCase):
         """
         request = self.request_factory.get('/somewhere/')
         response = get_view(request)
-
-        self.assertEqual(response.status_code, 200)
         self.assertContains(response, 'This is a test')
 
     def test_trace_request_from_factory(self):
@@ -739,6 +917,4 @@ class RequestFactoryTest(SimpleTestCase):
         response = trace_view(request)
         protocol = request.META["SERVER_PROTOCOL"]
         echoed_request_line = "TRACE {} {}".format(url_path, protocol)
-
-        self.assertEqual(response.status_code, 200)
         self.assertContains(response, echoed_request_line)
